// File_Jpeg - Info for NewFormat files
// Copyright (C) 2005-2007 Jerome Martinez, Zen@MediaArea.net
//
// This library is free software: you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this library. If not, see <http://www.gnu.org/licenses/>.
//
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Links
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//
// http://www.fileformat.info/format/jpeg/
// http://park2.wakwak.com/~tsuruzoh/Computer/Digicams/exif-e.html
// http://www.w3.org/Graphics/JPEG/jfif3.pdf
// http://www.sentex.net/~mwandel/jhead/
//
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//---------------------------------------------------------------------------
// Compilation conditions
#include <MediaInfo/Setup.h>
#ifdef __BORLANDC__
    #pragma hdrstop
#endif
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
#if defined(MEDIAINFO_JPEG_YES)
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
#include "MediaInfo/Image/File_Jpeg.h"
#include "ZenLib/Utils.h"
using namespace ZenLib;
//---------------------------------------------------------------------------

namespace MediaInfoLib
{

//***************************************************************************
// Format
//***************************************************************************

//---------------------------------------------------------------------------
File_Jpeg::File_Jpeg()
: File__Analyze()
{
    MustParseTheHeaderFile=true;
}

//***************************************************************************
// Buffer
//***************************************************************************

//---------------------------------------------------------------------------
void File_Jpeg::FileHeader_Parse()
{
    //Parsing
    Element_Begin("Header", 10);
    int16u identifier;
    Get_B2 (identifier,                                         "identifier");
    Element_End();

    FILLING_BEGIN();
        if (identifier!=0xFFD8)
        {
            Finnished();
            return;
        }
    FILLING_END();
}

//---------------------------------------------------------------------------
// An Element
// Size                             2 bytes, Pos=4
// Code                             5 bytes, Pos=6
//
void File_Jpeg::Header_Parse()
{
    //Parsing
    int16u code, size;
    Get_B2 (code,                                               "code");
    Get_B2 (size,                                               "size");

    //Filling
    Header_Fill_Code(code, Ztring().From_CC2(code));
    Header_Fill_Size(2+size);
}

//---------------------------------------------------------------------------
// Element parse
//
void File_Jpeg::Data_Parse()
{
    //Parsing
    switch (Element_Code)
    {
        case Jpeg::SOF0 : SOF0(); break;
        case Jpeg::SOF1 : SOF1(); break;
        case Jpeg::SOF2 : SOF2(); break;
        case Jpeg::SOF3 : SOF3(); break;
        case Jpeg::SOF9 : SOF9(); break;
        case Jpeg::SOFA : SOFA(); break;
        case Jpeg::SOFB : SOFB(); break;
        case Jpeg::DHT  : DHT (); break;
        case Jpeg::SOI  : SOI (); break;
        case Jpeg::EOI  : EOI (); break;
        case Jpeg::DQT  : DQT (); break;
        case Jpeg::SOS  : SOS (); break;
        case Jpeg::APP0 : APP0(); break;
        case Jpeg::APP1 : APP1(); break;
        default : Skip_XX(Element_Size,                         "Data");
    }
}

//***************************************************************************
// Elements
//***************************************************************************

//---------------------------------------------------------------------------
void File_Jpeg::SOF_()
{
    //Parsing
    int16u Height, Width;
    int8u  Resolution, Count;
    Get_B1 (Resolution,                                         "Resolution");
    Get_B2 (Height,                                             "Height");
    Get_B2 (Width,                                              "Width");
    Get_B1 (Count,                                              "Number of image components in frame");
    for (int8u Pos=0; Pos<Count; Pos++)
    {
        Skip_B1(                                                "Identifier");
        Skip_B1(                                                "sampling factor");
        Skip_B1(                                                "Quantization table destination selector");
    }

    FILLING_BEGIN();
        Stream_Prepare(Stream_General);
        Fill("Format", "JPEG");
        Stream_Prepare(Stream_Image);
        Fill("Codec", "JPEG");
        Fill("Codec/String", "JPEG"); //To Avoid automatic filling
        Fill("Resolution", Resolution);
        Fill("Height", Height);
        Fill("Width", Width);
    FILLING_END();
}

//---------------------------------------------------------------------------
void File_Jpeg::SOF0()
{
    Element_Info("Baseline DCT");
    SOF_();
}

//---------------------------------------------------------------------------
void File_Jpeg::SOF1()
{
    Element_Info("Extended sequential DCT, Huffman coding");
    SOF_();
}

//---------------------------------------------------------------------------
void File_Jpeg::SOF2()
{
    Element_Info("Progressive DCT, Huffman coding");
    SOF_();
}

//---------------------------------------------------------------------------
void File_Jpeg::SOF3()
{
    Element_Info("Lossless (sequential), Huffman coding");
    SOF_();
}

//---------------------------------------------------------------------------
void File_Jpeg::SOF9()
{
    Element_Info("Extended sequential DCT, arithmetic coding");
    SOF_();
}

//---------------------------------------------------------------------------
void File_Jpeg::SOFA()
{
    Element_Info("Progressive DCT, arithmetic coding");
    SOF_();
}

//---------------------------------------------------------------------------
void File_Jpeg::SOFB()
{
    Element_Info("Lossless (sequential), arithmetic coding");
    SOF_();
}

//---------------------------------------------------------------------------
void File_Jpeg::DHT()
{
    Element_Info("Define Huffman table(s)");

    //Parsing
    Skip_XX(Element_Size,                                       "Data");
}

//---------------------------------------------------------------------------
void File_Jpeg::SOI()
{
    Element_Info("Start Of Image");

    //Parsing
    Skip_XX(Element_Size,                                       "Data");
}

//---------------------------------------------------------------------------
void File_Jpeg::EOI()
{
    Element_Info("End Of Image");

    //Parsing
    Skip_XX(Element_Size,                                       "Data");
}

//---------------------------------------------------------------------------
void File_Jpeg::SOS()
{
    Element_Info("Start Of Scan");

    //Parsing
    int8u Count;
    Get_B1 (Count,                                              "Number of image components in scan");
    for (int8u Pos=0; Pos<Count; Pos++)
    {
        Skip_B1(                                                "Scan component selector");
        Skip_B1(                                                "Entropy coding table destination selector");
    }
    Skip_B1(                                                    "Start of spectral or predictor selection");
    Skip_B1(                                                    "End of spectral selection");
    Skip_B1(                                                    "Successive approximation bit position");

    //Filling
    Finnished(); //No need of more
}

//---------------------------------------------------------------------------
void File_Jpeg::DQT()
{
    Element_Info("Define quantization table(s)");

    //Parsing
    Skip_XX(Element_Size,                                       "Data");
}

//---------------------------------------------------------------------------
void File_Jpeg::APP0()
{
    Element_Info("Application Segment 0");

    //Parsing
    int64u Name;
    Get_C5(Name,                                                "Name");
    if (Name==CC5("JFIF\0"))
        APP0_JFIF();
    if (Name==CC5("JFFF\0"))
        APP0_JFXX();
}

//---------------------------------------------------------------------------
void File_Jpeg::APP0_JFIF()
{
    //Parsing
    Element_Begin("JFIF");
        int16u Width, Height;
        int8u  Unit, ThumbailX, ThumbailY;
        Skip_B2(                                                "Version");
        Get_B1 (Unit,                                           "Unit"); //0=Pixels, 1=dpi, 2=dpcm
        Get_B2 (Width,                                          "Xdensity");
        Get_B2 (Height,                                         "Ydensity");
        Get_B1 (ThumbailX,                                      "Xthumbail");
        Get_B1 (ThumbailY,                                      "Ythumbail");
        Skip_XX(3*ThumbailX*ThumbailY,                          "RGB Thumbail");
    Element_End();
}

//---------------------------------------------------------------------------
void File_Jpeg::APP0_JFXX()
{
    Element_Begin("Extension");
        Skip_B1(                                                "extension_code"); //0x10 Thumbnail coded using JPEG, 0x11 Thumbnail stored using 1 byte/pixel, 0x13 Thumbnail stored using 3 bytes/pixel
        if (Element_Size>Element_Offset)
            Skip_XX(Element_Size-Element_Offset,                "extension_data");
    Element_End();
}

//---------------------------------------------------------------------------
void File_Jpeg::APP0_JFXX_JPEG()
{
    //Parsing
    Element_Begin("Thumbail JPEG");
        if (Element_Size>Element_Offset)
            Skip_XX(Element_Size-Element_Offset,                "Data");
    Element_End();
}

//---------------------------------------------------------------------------
void File_Jpeg::APP0_JFXX_1B()
{
    //Parsing
    Element_Begin("Thumbail 1 byte per pixel");
        int8u  ThumbailX, ThumbailY;
        Get_B1 (ThumbailX,                                      "Xthumbail");
        Get_B1 (ThumbailY,                                      "Ythumbail");
        Skip_XX(768,                                            "Palette");
        Skip_XX(ThumbailX*ThumbailY,                            "Thumbail");
    Element_End();
}

//---------------------------------------------------------------------------
void File_Jpeg::APP0_JFXX_3B()
{
    //Parsing
    Element_Begin("Thumbail 3 bytes per pixel");
        int8u  ThumbailX, ThumbailY;
        Get_B1 (ThumbailX,                                      "Xthumbail");
        Get_B1 (ThumbailY,                                      "Ythumbail");
        Skip_XX(3*ThumbailX*ThumbailY,                          "RGB Thumbail");
    Element_End();
}

//---------------------------------------------------------------------------
void File_Jpeg::APP1()
{
    Element_Info("Application Segment 1");

    //Parsing
    int64u Name;
    Get_C6(Name,                                                "Name");

    if (Name==CC6("Exif\0\0"))
        APP1_EXIF();
}

//---------------------------------------------------------------------------
void File_Jpeg::APP1_EXIF()
{
    //Parsing
    Element_Begin("Exif");
        int32u Aligment;
        Get_C4(Aligment,                                        "Aligment");
        if (Aligment==0x49492A00)
            Skip_B4(                                            "First_IFD");
        if (Aligment==0x4D4D2A00)
            Skip_L4(                                            "First_IFD");
    Element_End();
}

//***************************************************************************
// Information
//***************************************************************************

//---------------------------------------------------------------------------
void File_Jpeg::HowTo(stream_t StreamKind)
{
         if (StreamKind==Stream_General)
    {
        General[0](_T("Format"), Info_HowTo)=_T("R");
    }
    else if (StreamKind==Stream_Image)
    {
        Video[0](_T("Codec"), Info_HowTo)=_T("Raw");
    }

    return;
}

} //NameSpace

#endif
